/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL v2.
You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */

#pragma once
#include "execution_defs.h"
#include "execution_manager.h"
#include "executor_abstract.h"
#include "index/ix.h"
#include "system/sm.h"

class SortExecutor : public AbstractExecutor {
   private:
    std::unique_ptr<AbstractExecutor> prev_;
    std::vector<ColMeta> cols_;                              // 框架中只支持一个键排序，需要自行修改数据结构支持多个键排序
    size_t tuple_num;
    bool is_desc_;
    std::vector<size_t> used_tuple;
    std::unique_ptr<RmRecord> current_tuple;
    std::vector<std::unique_ptr<RmRecord>> tuples;
    int local;
    int init_segment_len,result_len;
    std::string sort_file1_name;
    std::string sort_file2_name;
    int tempfd1,tempfd2;
    int len;
    int fd_rcd_num;
    std::unique_ptr<DiskManager> disk_manager;
   public:
    SortExecutor(std::unique_ptr<AbstractExecutor> prev, std::vector<TabCol> sel_cols, bool is_desc) {
        disk_manager = std::make_unique<DiskManager>();
        prev_ = std::move(prev);
        for(auto it : sel_cols){
            cols_.push_back(prev_->get_col_offset(it));
        }
        is_desc_ = is_desc;
        tuple_num = 0;
        used_tuple.clear();
        local = 0;
        init_segment_len = 100;
        result_len = 100;
        len = prev_->tupleLen();
        fd_rcd_num = 0;
    }
    //自定义比较函数
    bool rd_compare(std::unique_ptr<RmRecord> &a, std::unique_ptr<RmRecord> &b){
        for(auto it : cols_){
            int len = it.len;
            int offset = it.offset;
            char *str1 = a->data + offset;
            char *str2 = b->data + offset;
            if(it.type == TYPE_FLOAT){
                if(*(float *)str1 < *(float *)str2)
                    return true;
                else if(*(float *)str1 > *(float *)str2){
                    return false;
                }
            }
            else if(it.type == TYPE_INT){
                if(*(int *)str1 < *(int *)str2)
                    return true;
                else if(*(int *)str1 > *(int *)str2){
                    return false;
                }
            }
            else{
                int flag = memcmp(str1, str2, len);
                if(flag < 0){
                    return true;
                }
                else if(flag > 0){
                    return false;
                }
            }
        }
        return true;
    }
    bool rd_compare(char* a, char* b){
        for(auto it : cols_){
            int len = it.len;
            int offset = it.offset;
            char *str1 = a + offset;
            char *str2 = b + offset;
            if(it.type == TYPE_FLOAT){
                if(*(float *)str1 < *(float *)str2)
                    return true;
                else if(*(float *)str1 > *(float *)str2){
                    return false;
                }
            }
            else if(it.type == TYPE_INT){
                if(*(int *)str1 < *(int *)str2)
                    return true;
                else if(*(int *)str1 > *(int *)str2){
                    return false;
                }
            }
            else{
                int flag = memcmp(str1, str2, len);
                if(flag < 0){
                    return true;
                }
                else if(flag > 0){
                    return false;
                }
            }
        }
        return true;
    }
    void result_writing(int fd, std::vector<char*> &result, int len){
        char answer[len*result.size()];
        for(int i = 0;i < result.size(); i++){
            memcpy(answer + i * len, result[i], len);
        }
        ssize_t bytes_write = write(fd, answer, len*result.size());
        if(bytes_write != len*result.size()){
            throw UnixError();
        }
    } 
    void result_delete(std::vector<char*> &result){
        for(auto &elem : result) {
            delete[] elem;
        }
        result.clear();
    }
    std::string getCurrentTimeStr() {
        // 获取当前时间点
        auto now = std::chrono::system_clock::now();
        
        // 转换为系统时间（用于输出）
        std::time_t now_c = std::chrono::system_clock::to_time_t(now);
        std::tm* now_tm = std::localtime(&now_c);
        
        // 获取当前时间的毫秒部分
        auto ms = std::chrono::duration_cast<std::chrono::milliseconds>(now.time_since_epoch()) % 1000;
        
        // 格式化时间为字符串
        char buffer[30];
        snprintf(buffer, sizeof(buffer), "%04d-%02d-%02d %02d:%02d:%02d.%03lld",
                now_tm->tm_year + 1900, now_tm->tm_mon + 1, now_tm->tm_mday,
                now_tm->tm_hour, now_tm->tm_min, now_tm->tm_sec,
                static_cast<long long>(ms.count()));
        
        return std::string(buffer);
    }
    //从扫描算子中得到所有需要输出的元组并排序
    void beginTuple() override {
        std::string time = getCurrentTimeStr();
        sort_file1_name = time + "sorted_file1.txt";
        sort_file2_name = time + "sorted_file2.txt";
        disk_manager->create_file(sort_file1_name);
        disk_manager->create_file(sort_file2_name);
        tempfd1 = disk_manager->open_file(sort_file1_name);
        tempfd2 = disk_manager->open_file(sort_file2_name);

        std::vector<std::unique_ptr<RmRecord>> segment1;
        lseek(tempfd1, 0,SEEK_SET);
        int segment_len = init_segment_len;
        for(prev_->beginTuple(); !prev_->is_end(); prev_->nextTuple()){
            
            auto Tuple = prev_->Next();
            if (Tuple == nullptr)
                continue;

            segment1.push_back(move(Tuple));
            
            if(segment1.size() == segment_len){
                std::sort(segment1.begin(),segment1.end(),[this](std::unique_ptr<RmRecord> &a, std::unique_ptr<RmRecord> &b){
                    return rd_compare(a, b);
                    });
                //将每段排序完成的记录写入文件
                //组合结果，一次性写入
                char answer[len*segment_len];
                for(int i = 0;i < segment_len; i++){
                    memcpy(answer + i * len, segment1[i]->data, len);
                }
                write(tempfd1, answer, len*segment_len);
                fd_rcd_num += segment_len;
                segment1.clear();
            }
        }

        if(segment1.size() > 0){
            std::sort(segment1.begin(),segment1.end(),[this](std::unique_ptr<RmRecord> &a, std::unique_ptr<RmRecord> &b){
                    return rd_compare(a, b);
                    });
            char answer[len*segment1.size()];
            for(int i = 0;i < segment1.size(); i++){
                memcpy(answer + i * len, segment1[i]->data, len);
            }
            write(tempfd1, answer, len*segment1.size());

            fd_rcd_num+=segment1.size();
            segment1.clear();
        }
        
        while(fd_rcd_num > segment_len){
            //定位结果文件的文件头，接下来不再重复定位
            lseek(tempfd2, 0, SEEK_SET);
            //计算归并段数量
            int num,group;
            if(fd_rcd_num % segment_len == 0){
                num = fd_rcd_num/segment_len;
            }
            else{
                num = fd_rcd_num/segment_len + 1;
            }
            //计算归并组数量，由于归并段可能是奇数，那么最后一段最后再特殊处理
            group = num/2;
            //开始针对每一组归并
            for(int i = 0;i < group; i++){
                //设置基线以及段内偏移量
                int base1 = segment_len * 2 * i;
                int base2 = base1 + segment_len;
                int flag1 = 0, flag2 =0, right_len;

                //char result[result_len][llen];
                std::vector<char*> result;
                char rcd_group_1[len * init_segment_len], rcd_group_2[len * init_segment_len];
                //初始化两个归并组
                lseek(tempfd1, base1 * len, SEEK_SET);
                read(tempfd1, rcd_group_1, len * init_segment_len);
                flag1 += init_segment_len;
                lseek(tempfd1, base2 * len, SEEK_SET);
                //第二组中可能存在空位，因此每次整段读取的时候都必须检查剩余记录数能否把归并段填满
                if(base2 + init_segment_len > fd_rcd_num){
                    read(tempfd1, rcd_group_2, len * (fd_rcd_num - base2));
                    right_len = fd_rcd_num - base2;
                }
                else{
                    read(tempfd1, rcd_group_2, len * init_segment_len);
                    right_len = init_segment_len;
                }
                flag2 += init_segment_len;
                int seg1_ptr = 0, seg2_ptr = 0;
                while(true){
                    result.push_back(new char[len]);
                    if(rd_compare(rcd_group_1 + seg1_ptr*len, rcd_group_2 + seg2_ptr*len)) {
                        memcpy(result.back(), rcd_group_1 + seg1_ptr*len, len);
                        if(result.size() == result_len){
                            result_writing(tempfd2, result, len);
                            result_delete(result);
                        }
                        seg1_ptr++;

                        //左边的归并段读完了,接下来，要么读入新的归并段，要不把剩下的元组处理完
                        if(seg1_ptr >= init_segment_len){
                            //此时还可以读入新的归并段
                            if(flag1 < segment_len){
                                lseek(tempfd1, (base1 + flag1) * len, SEEK_SET);
                                read(tempfd1, rcd_group_1, len * init_segment_len);
                                flag1 += init_segment_len;
                                seg1_ptr = 0;
                                continue;
                            }
                            else{
                                //处理result中可能存在的记录
                                if(result.size() > 0){
                                    result_writing(tempfd2, result, len);
                                    result_delete(result);
                                }
                                
                                write(tempfd2, rcd_group_2+seg2_ptr*len, len*(right_len - seg2_ptr));

                                //处理右边归并段中的记录
                                lseek(tempfd1, (base2 + flag2) * len, SEEK_SET);
                                
                                while(flag2 < segment_len && fd_rcd_num-(base2+flag2) > 0){
                                    if(base2 +flag2 + init_segment_len > fd_rcd_num){
                                        right_len = fd_rcd_num-(base2+flag2);
                                    }
                                    else{
                                        right_len = init_segment_len;
                                    }
                                    read(tempfd1, rcd_group_2, len*right_len);
                                    write(tempfd2, rcd_group_2, len*right_len);
                                    flag2 += init_segment_len;
                                    //seg2_ptr = 0;
                                }
                                
                                break;
                            }
                        }  
                    }
                    else{
                        memcpy(result.back(), rcd_group_2 + seg2_ptr*len, len);
                        if(result.size() == result_len){
                            //lseek(tempfd2, len * )
                            result_writing(tempfd2, result, len);
                            result_delete(result);
                        }
                        seg2_ptr++;
                        //右边的归并段读完了,接下来，要么读入新的归并段，要不把剩下的元组处理完
                        if(seg2_ptr >= right_len){
                            //判断该段是不是右侧归并段的最后一段
                            if(flag2 < segment_len && fd_rcd_num-(base2+flag2) > 0){
                                lseek(tempfd1, (base2 + flag2) * len, SEEK_SET);
                                if(base2 +flag2 + init_segment_len > fd_rcd_num){
                                    right_len = fd_rcd_num-(base2+flag2);
                                    read(tempfd1, rcd_group_2, len*right_len);
                                }
                                else{
                                    read(tempfd1, rcd_group_2, len * init_segment_len);
                                    right_len = init_segment_len;
                                }
                                flag2 += init_segment_len;
                                seg2_ptr = 0;
                                continue;
                            }
                            else{
                                if(result.size() > 0){
                                    result_writing(tempfd2, result, len);
                                    result_delete(result);
                                }
                                write(tempfd2, rcd_group_1+seg1_ptr*len, (init_segment_len-seg1_ptr)*len);

                                lseek(tempfd1, (base1 + flag1) * len, SEEK_SET);
                                //char last_rcd[len * init_segment_len];
                                while(flag1 < segment_len){
                                    read(tempfd1, rcd_group_1, len * init_segment_len);
                                    write(tempfd2, rcd_group_1, len * init_segment_len);
                                    flag1 += init_segment_len;
                                }
                                break;
                            }
                        }
                    } 
                }
            }
            //开始处理最后可能独立的归并段
            if(num%2 != 0){
                // for(int i = (num - 1)*segment_len; i < fd_rcd_num; i++){

                // }
                lseek(tempfd1, segment_len*2*group*len, SEEK_SET);
                int left_num = fd_rcd_num - segment_len*2*group;
                char rcd[len*result_len];
                for(int i = 0;i < left_num/result_len; i++){
                    read(tempfd1, rcd, len*result_len);
                    write(tempfd2, rcd, len*result_len);
                }
                if(left_num%result_len > 0){
                    read(tempfd1, rcd, len*(left_num%result_len));
                    write(tempfd2, rcd, len*(left_num%result_len));
                }
            }
            //将结果文件的内容写入磁盘
            fsync(tempfd2);
            //交换文件句柄
            int t = tempfd1;
            tempfd1 = tempfd2;
            tempfd2 = t;
            //归并段长翻倍
            segment_len *= 2;
        }

        lseek(tempfd1, 0, SEEK_SET);
    }

    void nextTuple() override {
        local++;
    }

    std::unique_ptr<RmRecord> Next() override {
        std::unique_ptr<RmRecord> tuple = std::make_unique<RmRecord>(len);
        char rcd[len];
        read(tempfd1, rcd, len);
        memcpy(tuple->data, rcd, len);
        return std::move(tuple);
    }

    bool is_end() const override{
        if(local >= fd_rcd_num){
            //unlink(sort_file1_name.c_str());
            //unlink(sort_file2_name.c_str());
            disk_manager->close_file(tempfd1);
            disk_manager->close_file(tempfd2);
            disk_manager->destroy_file(sort_file1_name);
            disk_manager->destroy_file(sort_file2_name);
            return true;
        }
        else{
            return false;
        }
    }
    // 实现cols方法，上层需要调用
    const std::vector<ColMeta> &cols() const
    {
        return prev_->cols();
    };
    
    Rid &rid() override { return _abstract_rid; }
};