/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL v2.
You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */

#include "log_recovery.h"

/**
 * @description: analyze阶段，需要获得脏页表（DPT）和未完成的事务列表（ATT）
 */
void RecoveryManager::analyze() {
    buffer_.offset_ = 0;
    int log_file_offset;
    int read_len;
    //获取最后一个检查点的位置
    int fd = open("restart_file",O_CREAT,S_IRUSR|S_IWUSR|S_IRGRP|S_IROTH);
    struct stat stat_buf;
    int rc = stat("restart_file", &stat_buf);
    if(rc != 0){
        throw InternalError("restart_file_open_failure");
    }
    int restart_file_len = stat_buf.st_size;
    
    assert(restart_file_len % sizeof(int) == 0);
    //说明此时没建检查点
    if(restart_file_len == 0){
        log_file_offset = 0;
    }
    else{
        lseek(fd, restart_file_len - sizeof(int), SEEK_SET);
        read(fd, &log_file_offset, sizeof(int));
    }

    close(fd);
    do{
        read_len = disk_manager_->read_log(buffer_.buffer_, LOG_BUFFER_SIZE, log_file_offset);
        int buffer_offset = 0;
        while(buffer_offset + OFFSET_LOG_TOT_LEN + sizeof(uint32_t) < read_len){
            int log_len = *(uint32_t *)(buffer_.buffer_ + buffer_offset + OFFSET_LOG_TOT_LEN);
            //assert(buffer_offset + log_len <= read_len);
            if(buffer_offset + log_len > read_len){
                break;
            }
            LogType log_type = *(LogType*)(buffer_.buffer_+buffer_offset+OFFSET_LOG_TYPE);
            switch (log_type)
            {
            case LogType::begin:{
                BeginLogRecord * log_rcd = new BeginLogRecord();
                log_rcd->deserialize(buffer_.buffer_ + buffer_offset);
                lsn_offset_map[log_rcd->lsn_] = log_file_offset;
                lsn_len_map[log_rcd->lsn_] = log_rcd->log_tot_len_;

                //维护日志文件偏移信息和日志缓冲器内偏移信息
                log_file_offset += log_rcd->log_tot_len_;
                buffer_offset += log_rcd->log_tot_len_;

                //undo_txn_latest_lsn_map[log_rcd->log_tid_] = log_rcd->lsn_;
                //加入undo_list
                undo_list.push_back(std::make_pair(log_rcd->log_tid_, log_rcd->lsn_));

                delete log_rcd;
                break;
            }
            case LogType::commit:{
                CommitLogRecord * log_rcd = new CommitLogRecord();
                log_rcd->deserialize(buffer_.buffer_ + buffer_offset);
                lsn_offset_map[log_rcd->lsn_] = log_file_offset;
                lsn_len_map[log_rcd->lsn_] = log_rcd->log_tot_len_;

                //维护日志文件偏移信息和日志缓冲器内偏移信息
                log_file_offset += log_rcd->log_tot_len_;
                buffer_offset += log_rcd->log_tot_len_;

                //从undo_list中删除该事务的所有日志记录
                //undo_txn_latest_lsn_map.erase(log_rcd->log_tid_);
                txn_id_t target_txn_id = log_rcd->log_tid_;
                auto new_end = std::remove_if(undo_list.begin(), undo_list.end(),
                        [target_txn_id](const std::pair<txn_id_t, lsn_t>& elem) {
                            return elem.first == target_txn_id;
                        });
                undo_list.erase(new_end, undo_list.end());


                delete log_rcd;
                break;
            }
            case LogType::ABORT:{
                AbortLogRecord * log_rcd = new AbortLogRecord();
                log_rcd->deserialize(buffer_.buffer_ + buffer_offset);
                lsn_offset_map[log_rcd->lsn_] = log_file_offset;
                lsn_len_map[log_rcd->lsn_] = log_rcd->log_tot_len_;

                //维护日志文件偏移信息和日志缓冲器内偏移信息
                log_file_offset += log_rcd->log_tot_len_;
                buffer_offset += log_rcd->log_tot_len_;

                //从undo_list中删除该事务的所有日志记录
                //undo_txn_latest_lsn_map.erase(log_rcd->log_tid_);
                txn_id_t target_txn_id = log_rcd->log_tid_;
                auto new_end = std::remove_if(undo_list.begin(), undo_list.end(),
                        [target_txn_id](const std::pair<txn_id_t, lsn_t>& elem) {
                            return elem.first == target_txn_id;
                        });
                undo_list.erase(new_end, undo_list.end());

                delete log_rcd;
                break;
            }
            case LogType::INSERT:{
                InsertLogRecord * log_rcd = new InsertLogRecord();
                log_rcd->deserialize(buffer_.buffer_ + buffer_offset);
                lsn_offset_map[log_rcd->lsn_] = log_file_offset;
                lsn_len_map[log_rcd->lsn_] = log_rcd->log_tot_len_;

                //维护日志文件偏移信息和日志缓冲器内偏移信息
                log_file_offset += log_rcd->log_tot_len_;
                buffer_offset += log_rcd->log_tot_len_;

                //加入undo_list
                //undo_txn_latest_lsn_map[log_rcd->log_tid_] = log_rcd->lsn_;
                undo_list.push_back(std::make_pair(log_rcd->log_tid_, log_rcd->lsn_));
                redo_list.emplace_back(log_rcd->lsn_);

                delete log_rcd;
                break;
            }
            case LogType::UPDATE:
            {
                UpdateLogRecord * log_rcd = new UpdateLogRecord();
                log_rcd->deserialize(buffer_.buffer_ + buffer_offset);
                lsn_offset_map[log_rcd->lsn_] = log_file_offset;
                lsn_len_map[log_rcd->lsn_] = log_rcd->log_tot_len_;

                //维护日志文件偏移信息和日志缓冲器内偏移信息
                log_file_offset += log_rcd->log_tot_len_;
                buffer_offset += log_rcd->log_tot_len_;

                //加入undo_list
                //undo_txn_latest_lsn_map[log_rcd->log_tid_] = log_rcd->lsn_;
                undo_list.push_back(std::make_pair(log_rcd->log_tid_, log_rcd->lsn_));
                redo_list.emplace_back(log_rcd->lsn_);
                
                delete log_rcd;
                break;
            }
            case LogType::DELETE:
            {
                DeleteLogRecord * log_rcd = new DeleteLogRecord();
                log_rcd->deserialize(buffer_.buffer_ + buffer_offset);
                lsn_offset_map[log_rcd->lsn_] = log_file_offset;
                lsn_len_map[log_rcd->lsn_] = log_rcd->log_tot_len_;

                //维护日志文件偏移信息和日志缓冲器内偏移信息
                log_file_offset += log_rcd->log_tot_len_;
                buffer_offset += log_rcd->log_tot_len_;

                //加入undo_list
                //undo_txn_latest_lsn_map[log_rcd->log_tid_] = log_rcd->lsn_;
                undo_list.push_back(std::make_pair(log_rcd->log_tid_, log_rcd->lsn_));
                redo_list.emplace_back(log_rcd->lsn_);
                
                delete log_rcd;
                break;
            }
            default:
                throw InternalError("log_type_error");
                break;
            }
        }
    }while (read_len == LOG_BUFFER_SIZE);

    //由于删除undo_list中的元素导致undo_list中的元素失序，因此重新排序
    std::sort(undo_list.begin(), undo_list.end(),
        [](const std::pair<txn_id_t, lsn_t>& a, const std::pair<txn_id_t, lsn_t>& b) {
            return a.second > b.second; // 降序比较 lsn_t
        });
}

/**
 * @description: 重做所有未落盘的操作
 */
void RecoveryManager::redo() {
    for(lsn_t lsn : redo_list){
        int log_len = lsn_len_map[lsn];
        int log_fd_offset = lsn_offset_map[lsn];
        char log[log_len];
        disk_manager_->read_log(log, log_len, log_fd_offset);
        LogType log_type = *(LogType*)(log+OFFSET_LOG_TYPE);
        switch (log_type)
        {
        case LogType::INSERT:{
            InsertLogRecord* log_rcd = new InsertLogRecord();
            log_rcd->deserialize(log);

            std::string tab_name(log_rcd->table_name_,log_rcd->table_name_size_);
            auto rm_file_hdl = sm_manager_->fhs_[tab_name].get();
            
            if(rm_file_hdl->get_file_hdr().num_pages <= log_rcd->rid_.page_no){
                throw InternalError("page miss");
            }
            // if(rm_file_hdl->get_file_hdr().num_pages <= log_rcd->rid_.page_no){
            //     page_id_t new_page_no;
            //     while(rm_file_hdl->get_file_hdr().num_pages <= log_rcd->rid_.page_no){
            //         new_page_no = rm_file_hdl->create_new_page_handle().page->get_page_id().page_no;
            //     }
            //     assert(new_page_no == log_rcd->rid_.page_no);
            //     rm_file_hdl->recover_insert_record(log_rcd->rid_,log_rcd->insert_value_.data);
            // }
            
            auto page_hdl = rm_file_hdl->fetch_page_handle(log_rcd->rid_.page_no);
            if(page_hdl.page->get_page_lsn() < log_rcd->lsn_){
                rm_file_hdl->recover_insert_record(log_rcd->rid_,log_rcd->insert_value_.data);
            }
            buffer_pool_manager_->unpin_page(page_hdl.page->get_page_id(),true);

            std::cout<<"redo insert "<<log_rcd->lsn_<<std::endl;
                    
            delete log_rcd;
            break;
        }
        case LogType::DELETE:{
            DeleteLogRecord* log_rcd = new DeleteLogRecord();
            log_rcd->deserialize(log);

            std::string tab_name(log_rcd->table_name_,log_rcd->table_name_size_);
            auto rm_file_hdl = sm_manager_->fhs_[tab_name].get();
            //检查该页是否存在，如果不存在，则返回错误
            if(rm_file_hdl->get_file_hdr().num_pages <= log_rcd->rid_.page_no){
                throw InternalError("page miss");
            }

            auto page_hdl = rm_file_hdl->fetch_page_handle(log_rcd->rid_.page_no);
            if(page_hdl.page->get_page_lsn() < log_rcd->lsn_){
                rm_file_hdl->recover_delete_record(log_rcd->rid_,nullptr);
            }
            buffer_pool_manager_->unpin_page(page_hdl.page->get_page_id(),true);

            std::cout<<"redo delete "<<log_rcd->lsn_<<std::endl;

            // 释放delete log rcd的内存
            delete log_rcd;
            break;
        }
        case LogType::UPDATE:{
            UpdateLogRecord* log_rcd = new UpdateLogRecord();
            log_rcd->deserialize(log);

            std::string tab_name(log_rcd->table_name_,log_rcd->table_name_size_);
            auto rm_file_hdl = sm_manager_->fhs_[tab_name].get();

            //检查该页是否存在，如果不存在，则返回错误
            if(rm_file_hdl->get_file_hdr().num_pages <= log_rcd->rid_.page_no){
                throw InternalError("page miss");
            }

            auto page_hdl = rm_file_hdl->fetch_page_handle(log_rcd->rid_.page_no);
            if(page_hdl.page->get_page_lsn() < log_rcd->lsn_){
                rm_file_hdl->update_record(log_rcd->rid_,log_rcd->update_value_.data,nullptr);
            }
            buffer_pool_manager_->unpin_page(page_hdl.page->get_page_id(),true);

            std::cout<<"redo update "<<log_rcd->lsn_<<std::endl;

            // 释放update log rcd的内存
            delete log_rcd;
            break;
        }

        default:
            break;
        }
    }
}

/**
 * @description: 回滚未完成的事务
 */
void RecoveryManager::undo() {
    for(auto txn2lsn : undo_list){
        lsn_t lsn = txn2lsn.second;
        int log_len = lsn_len_map[lsn];
        int log_fd_offset = lsn_offset_map[lsn];
        char log[log_len];
        disk_manager_->read_log(log, log_len, log_fd_offset);
        LogType log_type = *(LogType*)(log+OFFSET_LOG_TYPE);
        switch (log_type)
        {
            case LogType::begin:{
                BeginLogRecord* log_rcd = new BeginLogRecord();
                log_rcd->deserialize(log);
                //lsn = log_rcd->prev_lsn_;

                std::cout<<"undo begin "<<log_rcd->lsn_<<std::endl;
                break;
            }
            case LogType::INSERT:{
                InsertLogRecord* log_rcd = new InsertLogRecord();

                log_rcd->deserialize(log);
                std::string tab_name(log_rcd->table_name_,log_rcd->table_name_size_);
                auto rm_file_hdl = sm_manager_->fhs_[tab_name].get();
                rm_file_hdl->recover_delete_record(log_rcd->rid_,nullptr);
                //lsn = log_rcd->prev_lsn_;

                std::cout<<"undo insert "<<log_rcd->lsn_<<std::endl;
                break;

            }
            case LogType::DELETE:{
                DeleteLogRecord* log_rcd = new DeleteLogRecord();

                log_rcd->deserialize(log);
                std::string tab_name(log_rcd->table_name_,log_rcd->table_name_size_);
                auto rm_file_hdl = sm_manager_->fhs_[tab_name].get();
                rm_file_hdl->recover_insert_record(log_rcd->rid_,log_rcd->delete_value_.data);

                //lsn = log_rcd->prev_lsn_;

                std::cout<<"undo delete "<<log_rcd->lsn_<<std::endl;
                break;
            }
            case LogType::UPDATE:{
                UpdateLogRecord* log_rcd = new UpdateLogRecord();

                log_rcd->deserialize(log);

                // 先将table name转化为string
                std::string tab_name(log_rcd->table_name_,log_rcd->table_name_size_);
                auto rm_file_hdl = sm_manager_->fhs_[tab_name].get();
                rm_file_hdl->update_record(log_rcd->rid_, log_rcd->old_value_.data, nullptr);

                //lsn = log_rcd->prev_lsn_;

                std::cout<<"undo update "<<log_rcd->lsn_<<std::endl;
                break;
            }
            default:{
                throw InternalError("log type falure");
                break;
            }
        }
        
    }
}

void RecoveryManager::rebuild_index(){
    //考虑到目前只实现一个表

    for (auto entry : sm_manager_->db_.tabs_){
        for(auto x : entry.second.indexes){
            sm_manager_->drop_index(entry.first, x.cols, nullptr);
            
            //遍历这个索引的所有属性名
            std::vector<std::string> cols;
            for(auto it : x.cols){
                cols.push_back(it.name);
            }
            sm_manager_->create_index(entry.first, cols, nullptr);
        }
    }

    // std::pair<std::string, std::vector<std::vector<ColMeta>>> tablename = sm_manager_->return_tbname_and_cols();
    // std::string table = tablename.first;
    // std::vector<std::vector<ColMeta>> cols_vector = tablename.second;
    // for(auto it : cols_vector){
    //     sm_manager_->drop_index(table, it, nullptr);
    //     std::vector<std::string> cols;
    //     for(auto x : it){
    //         cols.push_back(x.name);
    //     }
    //     sm_manager_->create_index(table, cols, nullptr);
    // }
    
}