/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL v2.
You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */

#include "sm_manager.h"

#include <sys/stat.h>
#include <unistd.h>

#include <fstream>

#include <sys/mman.h>   // 用于内存映射

#include "index/ix.h"
#include "record/rm.h"
#include "record_printer.h"
#include "common/common.h"
/**
 * @description: 判断是否为一个文件夹
 * @return {bool} 返回是否为一个文件夹
 * @param {string&} db_name 数据库文件名称，与文件夹同名
 */
bool SmManager::is_dir(const std::string& db_name) {
    struct stat st;
    return stat(db_name.c_str(), &st) == 0 && S_ISDIR(st.st_mode);
}

/**
 * @description: 创建数据库，所有的数据库相关文件都放在数据库同名文件夹下
 * @param {string&} db_name 数据库名称
 */
void SmManager::create_db(const std::string& db_name) {
    if (is_dir(db_name)) {
        throw DatabaseExistsError(db_name);
    }
    //为数据库创建一个子目录
    std::string cmd = "mkdir " + db_name;
    if (system(cmd.c_str()) < 0) {  // 创建一个名为db_name的目录
        throw UnixError();
    }
    if (chdir(db_name.c_str()) < 0) {  // 进入名为db_name的目录
        throw UnixError();
    }
    //创建系统目录
    DbMeta *new_db = new DbMeta();
    new_db->name_ = db_name;

    // 注意，此处ofstream会在当前目录创建(如果没有此文件先创建)和打开一个名为DB_META_NAME的文件
    std::ofstream ofs(DB_META_NAME);

    // 将new_db中的信息，按照定义好的operator<<操作符，写入到ofs打开的DB_META_NAME文件中
    ofs << *new_db;  // 注意：此处重载了操作符<<

    delete new_db;

    // 创建日志文件
    disk_manager_->create_file(LOG_FILE_NAME);

    // 回到根目录
    if (chdir("..") < 0) {
        throw UnixError();
    }
}

/**
 * @description: 删除数据库，同时需要清空相关文件以及数据库同名文件夹
 * @param {string&} db_name 数据库名称，与文件夹同名
 */
void SmManager::drop_db(const std::string& db_name) {
    if (!is_dir(db_name)) {
        throw DatabaseNotFoundError(db_name);
    }
    std::string cmd = "rm -r " + db_name;
    if (system(cmd.c_str()) < 0) {
        throw UnixError();
    }
}

/**
 * @description: 打开数据库，找到数据库对应的文件夹，并加载数据库元数据和相关文件
 * @param {string&} db_name 数据库名称，与文件夹同名
 */
void SmManager::open_db(const std::string& db_name) {
    // 加载的数据都在DB_META_NAME文件中,格式如下:
    // 数据库名 表数量
    // 表名1 表名1的TabMeta
    // ...
    if (!is_dir(db_name)) {
        throw DatabaseNotFoundError(db_name);
    }
    if (chdir(db_name.c_str()) < 0) {  // 进入名为db_name的目录
        throw UnixError();
    }
    // 注意，此处ifstream会在当前目录打开一个名为DB_META_NAME的文件
    std::ifstream ifs(DB_META_NAME);
    if (!ifs) 
    {
        // 如果没有对应文件则抛出异常
        throw FileNotFoundError(DB_META_NAME);    
    }
    // 将DB_META_NAME中的信息，按照定义好的operator>>操作符，通过ifs写入SmManager下的db_中
    ifs >> db_;  // 注意：此处重载了操作符>> “sm_meta.h”
    // 目前DbMeta db_(当前打开的数据库的元数据)已经全部加载完毕
    // 继续加载std::unordered_map<std::string, std::unique_ptr<RmFileHandle>> fhs_;当前数据库中每张表的数据文件
    // std::unordered_map<std::string, std::unique_ptr<IxIndexHandle>> ihs_;当前数据库中每个索引的文件
    // 该数据库的各个表的数据信息都要读入
    for (auto &entry : db_.tabs_) 
    {   
        fhs_.emplace(entry.second.name, rm_manager_->open_file(entry.second.name)); //数据文件fhs_哈希表赋值
        TabMeta current_table = entry.second;
        for (auto &index : current_table.indexes) 
        {   
            // 索引文件ihs_哈希表赋值
            ihs_.emplace(ix_manager_->get_index_name(current_table.name, index.cols), ix_manager_->open_index(current_table.name, index.cols));
        }
    }
}

/**
 * @description: 把数据库相关的元数据刷入磁盘中
 */
void SmManager::flush_meta() {
    // 默认清空文件
    std::ofstream ofs(DB_META_NAME);
    ofs << db_;
}

/**
 * @description: 关闭数据库并把数据落盘
 */
void SmManager::close_db() {
    
}

/**
 * @description: 显示所有的表,通过测试需要将其结果写入到output.txt,详情看题目文档
 * @param {Context*} context 
 */
void SmManager::show_tables(Context* context) {
    std::fstream outfile;
    outfile.open("output.txt", std::ios::out | std::ios::app);
    outfile << "| Tables |\n";
    RecordPrinter printer(1);
    printer.print_separator(context);
    printer.print_record({"Tables"}, context);
    printer.print_separator(context);
    for (auto &entry : db_.tabs_) {
        auto &tab = entry.second;
        printer.print_record({tab.name}, context);
        outfile << "| " << tab.name << " |\n";
    }
    printer.print_separator(context);
    outfile.close();
}

/**
 * @description: 显示表的元数据
 * @param {string&} tab_name 表名称
 * @param {Context*} context 
 */
void SmManager::desc_table(const std::string& tab_name, Context* context) {
    TabMeta &tab = db_.get_table(tab_name);

    std::vector<std::string> captions = {"Field", "Type", "Index"};
    RecordPrinter printer(captions.size());
    // Print header
    printer.print_separator(context);
    printer.print_record(captions, context);
    printer.print_separator(context);
    // Print fields
    for (auto &col : tab.cols) {
        std::vector<std::string> field_info = {col.name, coltype2str(col.type), col.index ? "YES" : "NO"};
        printer.print_record(field_info, context);
    }
    // Print footer
    printer.print_separator(context);
}

/**
 * @description: 创建表
 * @param {string&} tab_name 表的名称
 * @param {vector<ColDef>&} col_defs 表的字段
 * @param {Context*} context 
 */
void SmManager::create_table(const std::string& tab_name, const std::vector<ColDef>& col_defs, Context* context) {
    if (db_.is_table(tab_name)) {
        throw TableExistsError(tab_name);
    }
    // Create table meta
    int curr_offset = 0;
    TabMeta tab;
    tab.name = tab_name;
    for (auto &col_def : col_defs) {
        ColMeta col = {.tab_name = tab_name,
                       .name = col_def.name,
                       .type = col_def.type,
                       .len = col_def.len,
                       .offset = curr_offset,
                       .index = false};
        curr_offset += col_def.len;
        tab.cols.push_back(col);
    }
    // Create & open record file
    int record_size = curr_offset;  // record_size就是col meta所占的大小（表的元数据也是以记录的形式进行存储的）
    rm_manager_->create_file(tab_name, record_size);
    db_.tabs_[tab_name] = tab;
    // fhs_[tab_name] = rm_manager_->open_file(tab_name);
    fhs_.emplace(tab_name, rm_manager_->open_file(tab_name));

    flush_meta();
}

/**
 * @description: 删除表
 * @param {string&} tab_name 表的名称
 * @param {Context*} context
 */
void SmManager::drop_table(const std::string& tab_name, Context* context) {
    if (!db_.is_table(tab_name)) {
        throw TableNotFoundError(tab_name);
    }
    auto it = fhs_.find(tab_name);
    if(it == fhs_.cend()){
        throw TableNotFoundError(tab_name);
    }
    std::unique_ptr<RmFileHandle>& value = fhs_[tab_name];
    RmFileHandle * fd = value.get();
    rm_manager_->close_file(fd);
    buffer_pool_manager_->delete_all_pages(fd->GetFd());
    rm_manager_->destroy_file(tab_name);
    
    fhs_.erase(tab_name);
    // lun 需要添加删除索引变量和索引文件相关的东西
    auto tabmeta = db_.tabs_[tab_name];
    for(auto it : tabmeta.indexes){
        drop_index(tab_name, it.cols, context);
    }
    db_.tabs_.erase(tab_name);
    flush_meta();
}

/**
 * @description: 创建索引
 * @param {string&} tab_name 表的名称
 * @param {vector<string>&} col_names 索引包含的字段名称
 * @param {Context*} context
 */
// 有bug，没有数据的时候创建索引不能创建文件lun
void SmManager::create_index(const std::string& tab_name, const std::vector<std::string>& col_names, Context* context) {
    // 检查是否创建索引的表名是否存在
    if (!(db_.is_table(tab_name))) {
        throw TableNotFoundError(tab_name);
    }
    // 表名合法后，检查这个表是否已经存在这个索引
    if (ix_manager_->exists(tab_name, col_names))
    {
        throw IndexExistsError(tab_name, col_names);
    }

    std::vector<ColMeta> index_cols;    //用来调用ix_manager_的create_index()函数
    TabMeta &tab = db_.get_table(tab_name); //用来为index_cols和TabMeta结构体中IndexMeta类型的indexes赋值
    // 创建 index meta 加入TabMeta结构体中的indexes成员中
    // struct IndexMeta {
    //      std::string tab_name;           // 索引所属表名称
    //      int col_tot_len;                // 索引字段长度总和
    //      int col_num;                    // 索引字段数量
    //      std::vector<ColMeta> cols;      // 索引包含的字段
    // }

    IndexMeta ind;
    ind.tab_name = tab_name;
    ind.col_tot_len = 0;
    ind.col_num = 0;
    for (auto &col_name : col_names) {
        // 检查表中创建索引的对应每列是否存在
        if (!(tab.is_col(col_name)))
        {
            throw ColumnNotFoundError(col_name);
        }
        else
        {   
            std::vector<ColMeta>::iterator iter = tab.get_col(col_name);
            ColMeta col = *iter; 
            index_cols.push_back(col); 
            // col.index------unused    没被使用，不用更改这个字段
            ind.col_num += 1;
            ind.col_tot_len += col.len;
            ind.cols.push_back(col);
        }
    }
    std::string index_file_name = ix_manager_->get_index_name(tab_name, col_names);
    ix_manager_->create_index(tab_name, index_cols);    //创建文件，初始化信息等等
    ihs_.emplace(index_file_name, ix_manager_->open_index(tab_name, col_names));   //哈希表赋值
    // db_.tabs_[tab_name].indexes.push_back(ind);
    tab.indexes.push_back(ind); //tab中索引赋值
    // 这里tab的数据应该是可以的覆盖的，因为用的是引用lun
    auto file_handle = fhs_[tab_name].get();
    for (RmScan scan(file_handle); !scan.is_end(); scan.next()) {
        char key[ind.col_tot_len];
        auto rec = file_handle->get_record(scan.rid(), context);
        int key_offset = 0;
        for(auto col : index_cols)
        {
            memcpy(key + key_offset, rec->data + col.offset, col.len);
            key_offset += col.len;
        }
        ihs_.at(index_file_name).get() -> insert_entry(key, scan.rid(), nullptr);  // lun后续需要修改tran 
    }
    flush_meta();
}

/**
 * @description: 删除索引
 * @param {string&} tab_name 表名称
 * @param {vector<string>&} col_names 索引包含的字段名称
 * @param {Context*} context
 */
void SmManager::drop_index(const std::string& tab_name, const std::vector<std::string>& col_names, Context* context) {
    // 检查是否删除索引的表名是否存在
    if (!(db_.is_table(tab_name))) {
        throw TableNotFoundError(tab_name);
    }
    // 表名合法后，检查这个表是否存在索引
    if (!(ix_manager_->exists(tab_name, col_names)))
    {
        throw IndexNotFoundError(tab_name, col_names);
    }
    // 如果上方检查出表存在该索引后，那么应该语句中的每列都存在，不用额外检查
    TabMeta &tab = db_.get_table(tab_name);

    std::string index_file_name = ix_manager_->get_index_name(tab_name, col_names);

    ix_manager_->close_index(ihs_.at(index_file_name).get());
    buffer_pool_manager_->delete_all_pages(ihs_.at(index_file_name).get()->get_fd());
    ix_manager_->destroy_index(tab_name, col_names);
    ihs_.erase(index_file_name);
    std::vector<IndexMeta>::iterator iter = tab.get_index_meta(col_names);
    db_.tabs_[tab_name].indexes.erase(iter);
    flush_meta();   //有什么用lun
}

/**
 * @description: 删除索引
 * @param {string&} tab_name 表名称
 * @param {vector<ColMeta>&} 索引包含的字段元数据
 * @param {Context*} context
 */
void SmManager::drop_index(const std::string& tab_name, const std::vector<ColMeta>& cols, Context* context) {
    // 检查是否删除索引的表名是否存在
    if (!(db_.is_table(tab_name))) {
        throw TableNotFoundError(tab_name);
    }
    // 表名合法后，检查这个表是否存在索引
    std::vector<std::string> col_names;
    for(auto col : cols)
    {
        col_names.push_back(col.name);
    }
    if (!(ix_manager_->exists(tab_name, col_names)))
    {
        throw IndexNotFoundError(tab_name, col_names);
    }
    // 如果上方检查出表存在该索引后，那么应该语句中的每列都存在，不用额外检查
    TabMeta &tab = db_.get_table(tab_name);

    std::string index_file_name = ix_manager_->get_index_name(tab_name, col_names);
    
    ix_manager_->close_index(ihs_.at(index_file_name).get());
    buffer_pool_manager_->delete_all_pages(ihs_.at(index_file_name).get()->get_fd());
    ix_manager_->destroy_index(tab_name, col_names);
    ihs_.erase(index_file_name);
    std::vector<IndexMeta>::iterator iter = tab.get_index_meta(col_names);
    db_.tabs_[tab_name].indexes.erase(iter);
    flush_meta();   //有什么用lun
}

/**
 * @description: 显示所有的唯一索引（因为题目只要求实现唯一索引）,通过测试需要将其结果写入到output.txt,详情看题目文档
 *               在output.txt中输出为评判格式，在客户端目前返回自定义的格式
 * @param {const std::string&} tab_name 查询的
 * @param {Context*} context 
 */
void SmManager::show_index(const std::string& tab_name, Context* context) {
    std::fstream outfile;   //期望输出： | 表名 | unique | (索引1)/(索引1,索引2,...) | 
    outfile.open("output.txt", std::ios::out | std::ios::app);  //打开后追加到上次写的最后，不会重起一行

    std::vector<std::string> captions = {"Table name", "Type", "Index"};
    RecordPrinter printer(captions.size());
    printer.print_separator(context);
    printer.print_record(captions, context);
    printer.print_separator(context);

    for (auto &entry : db_.tabs_[tab_name].indexes) 
    {   
        outfile << "| " << tab_name << " | " << "unique"<< " | ";
        std::string index_str = "(";  //用于output.txt输出和客户端输出
        bool isFirst = true; // 添加标记变量
        for (auto &col : entry.cols)
        {
            if (!isFirst) // 如果不是第一个元素，则输出逗号
            {   
                index_str.append(",");
            }
            index_str.append(col.name);
            isFirst = false; // 将标记设置为false，表示下一个将是非首个元素
        }
        index_str.append(")");
        std::vector<std::string> field_info = {tab_name, "unique", index_str};
        printer.print_record(field_info, context);

        outfile << index_str << " |\n"; //必须加/n，下次才能重起一行输入文件中
    }

    printer.print_separator(context);
    outfile.close();
}

std::pair<std::string, std::vector<std::vector<ColMeta>>> SmManager::return_tbname_and_cols(){
    std::pair<std::string, std::vector<std::vector<ColMeta>>> tablenames_and_cols;
    for (auto entry : db_.tabs_){
        tablenames_and_cols.first = entry.first;
        for(auto x : db_.tabs_[entry.first].indexes){
            tablenames_and_cols.second.push_back(x.cols);
        }
        break;
    }
    return tablenames_and_cols;
}

/**
 * @description: load————用于性能测试的语句————向表中插入.csv文件的数据
 * @param {string&} table_name 表的名称
 * @param {string&} file_name 读取文件的名称
 * @param {Context*} context 
 */
// void SmManager::load_table(const std::string& table_name, const std::string& file_name, Context* context) {
//         // std::vector<std::unique_ptr<RmRecord>> rows_value; // 用来接收文件中所有行所有列的数据
//         std::vector<ColMeta> all_cols;
//         all_cols = db_.get_table(table_name).cols;
//         RmFileHandle *fh_ = fhs_.at(table_name).get();
//         auto tab = db_.tabs_.at(table_name);
//         // 打开文件
//         std::ifstream file(file_name);
//         if (!file) {
//             std::cerr << "File not found." << std::endl;
//             throw FileNotFoundError(file_name);
//         }
//         // 跳过第一行列名的信息，从第二行开始读取
//         std::string line;
//         std::getline(file, line);
//         // 读取剩余的行
//         while (std::getline(file, line)) {
//             auto rec = std::make_unique<RmRecord>(fh_->get_file_hdr().record_size);
//             std::stringstream ss(line);
//             std::string item;   // 读取一行中单列的值为string类型
//             int col_num = 0;
//             // 按逗号分割每一行
//             while (std::getline(ss, item, ',')) {
//                 // 移除末尾的换行符
//                 if(!item.empty() && item.back() == '\r') {
//                     item.pop_back();
//                 }
//                 char* val[all_cols[col_num].len];
//                 // val.str_to_value(item, all_cols[col_num].type);
//                 switch (all_cols[col_num].type)
//                 {
//                     case ColType::TYPE_INT:{
//                         *(int *)val = std::stoi(item);
//                         break;
//                     }
//                     case ColType::TYPE_FLOAT:{
//                         *(float *)val = std::stof(item);  // 可能会有精度问题 lun
//                         break;
//                     }
//                     case ColType::TYPE_STRING:{
//                         memcpy(val, item.c_str(), item.size());
//                         break;   
//                     }
//                     default:
//                         break;
//                 }
//                 //val.init_raw(all_cols[col_num].len);
//                 memcpy(rec->data + all_cols[col_num].offset, val, all_cols[col_num].len);
//                 col_num++;
//             }
//             // 指向文件中一行的所有信息的指针放到rows_value中
//             // rows_value.push_back(rec);
//             Rid rid;
//             rid = fh_->insert_record(rec->data, context);
            
//             //索引维护
//             for(size_t i = 0; i < tab.indexes.size(); ++i){
//                 auto& index = tab.indexes[i];
//                 char *key = new char[index.col_tot_len];
//                 int key_offset = 0;
//                 for(size_t j = 0; j < index.col_num; j++){
//                     memcpy(key + key_offset, rec->data + index.cols[j].offset, index.cols[j].len);
//                     key_offset += index.cols[j].len; 
//                 }
//                 auto ih = ihs_.at(get_ix_manager()->get_index_name(table_name, index.cols)).get();
//                 ih->insert_entry(key, rid, context->txn_);
//                 delete[] key;
//             }
//         }
//         // 关闭文件
//         file.close();
// }
bool SmManager::load_table(const std::string& table_name, const std::string& file_name, Context* context) {

    // 使用命令wc -l file.csv得到行数
	std::string command = "wc -l " + file_name;
	FILE* pipe = popen(command.c_str(), "r");
	if (!pipe) {
		std::cerr << "Error executing command" << std::endl;
		return 1;
	}

	char buffer[128];
	std::string result = "";
	while (!feof(pipe)) {
		if (fgets(buffer, 128, pipe) != NULL)
			result += buffer;
	}

	pclose(pipe);

	// 获取 table
	auto& tab_ = db_.get_table(table_name);
	auto file_handle = fhs_.at(table_name).get();
    if(!tab_.indexes.empty())
    {
        auto& index_meta = tab_.indexes[0];
		std::vector<std::string> col_names;
		for (auto& col : index_meta.cols) {
			col_names.push_back(col.name);
		}

        // 获取所有索引字段的元数据
		std::vector<ColMeta> index_cols;
		for (auto it = col_names.begin(); it != col_names.end(); it++)
		{
			index_cols.push_back(*tab_.get_col(*it));
		}
        auto ih = ihs_.at(get_ix_manager()->get_index_name(table_name, index_cols)).get();

        // 计算索引字段长度总和
		int col_tot_len = 0;
		// int col_num = index_cols.size();
		for (ColMeta& col : index_cols)
		{
			col_tot_len += col.len;
		}

        int fd = open(file_name.c_str(), O_RDONLY);
        if (fd == -1) {
            perror("Error opening file");
            return 1;
        }

        // 获取文件大小
        off_t file_size = lseek(fd, 0, SEEK_END);

        // 将文件映射到内存
        char* file_memory = (char*)mmap(NULL, file_size, PROT_READ, MAP_PRIVATE, fd, 0);
        if (file_memory == MAP_FAILED) {
            perror("Error mapping file to memory");
            close(fd);
            return false;
        }

        char* s = file_memory, * t = file_memory;

        // 跳过header
        while (*t != '\n')t++;
        s = t + 1;
        t = s;

        size_t num_records_per_page = file_handle->get_file_hdr().num_records_per_page;
        size_t data_size = file_handle->get_file_hdr().record_size * num_records_per_page;
        char* head = new char[data_size];
        char* data = head;
        int record_num = 0;
        int page_num = 1;

        int col_idx = 0;
        for (; t < file_memory + file_size; t++) {
            if (*t == ',' || *t == '\n' || *t == '\r') {
                if (s == t) {
                    s++;
                    continue;
                }
                auto& col_meta = tab_.cols[col_idx++];
                switch (col_meta.type)
                {
                case TYPE_FLOAT:
                    *(double*)(data + col_meta.offset) = std::atof(s);
                    break;
                case TYPE_INT:
                    *(int*)(data + col_meta.offset) = atoi(s);
                    break;
                case TYPE_STRING: {
                    memcpy(data + col_meta.offset, s, t - s);
                    break;
                }
                default:
                    return false;
                }
                s = t + 1;
                if (*t == '\n' || *t == '\r') {
                    col_idx = 0;
                    // 插入数据
                    Rid ptr = { page_num, record_num % num_records_per_page };		// 从1，0开始
                    record_num++;

                    // 初始化索引字段的数据
                    char* key = new char[col_tot_len];
                    int offset = 0;
                    for (ColMeta& col : index_cols)
                    {
                        memcpy(key + offset, data + col.offset, col.len);
                        offset += col.len;
                    }
                    
                    ih->insert_entry(key, ptr, context->txn_);

                    if (record_num % num_records_per_page == 0 || t == file_memory + file_size - 1) {
                        // 批量插入一页
                        file_handle->load_record(head, record_num % num_records_per_page);
                        page_num++;
                        data = head;
                    }
                    else {
                        data += file_handle->get_file_hdr().record_size;
                    }
                }
            }
        }
        delete[] head;
        // 解除内存映射
        if (munmap(file_memory, file_size) == -1) {
            perror("Error unmapping memory");
        }
        close(fd);
    }
    else
    {
        int fd = open(file_name.c_str(), O_RDONLY);
		if (fd == -1) {
			perror("Error opening file");
			return 1;
		}

		// 获取文件大小
		off_t file_size = lseek(fd, 0, SEEK_END);

		// 将文件映射到内存
		char* file_memory = (char*)mmap(NULL, file_size, PROT_READ, MAP_PRIVATE, fd, 0);
		if (file_memory == MAP_FAILED) {
			perror("Error mapping file to memory");
			close(fd);
			return false;
		}

		char* s = file_memory, * t = file_memory;

		// 跳过header
		while (*t != '\n')t++;
		s = t + 1;
		t = s;

		size_t num_records_per_page = file_handle->get_file_hdr().num_records_per_page;
		size_t data_size = file_handle->get_file_hdr().record_size * num_records_per_page;
		char* head = new char[data_size];
		char* data = head;
		int record_num = 0;
		int page_num = 1;

		int col_idx = 0;
		for (; t < file_memory + file_size; t++) {
			if (*t == ',' || *t == '\n' || *t == '\r') {
				if (s == t) {
					s++;
					continue;
				}
				auto& col_meta = tab_.cols[col_idx++];
				switch (col_meta.type)
				{
                    case TYPE_FLOAT:
                        *(double*)(data + col_meta.offset) = std::atof(s);
                        break;
                    case TYPE_INT:
                        *(int*)(data + col_meta.offset) = atoi(s);
                        break;
                    case TYPE_STRING: {
                        memcpy(data + col_meta.offset, s, t - s);
                        break;
				}
				default:
					return false;
				}
				s = t + 1;
				if (*t == '\n' || *t == '\r') {
					col_idx = 0;

					Rid ptr = { page_num, record_num % num_records_per_page };		// 从1，0开始
					record_num++;
					if (record_num % num_records_per_page == 0 || t == file_memory + file_size - 1) {
						// 批量插入一页
						file_handle->load_record(head, record_num % num_records_per_page);
						page_num++;
						data = head;
						//memset(data, 0, data_size);
					}
					else {
						data += file_handle->get_file_hdr().record_size;
					}
					//memset(data, 0, file_handle->get_file_hdr().record_size);
				}
			}
		}
		delete[] head;
		// 解除内存映射
		if (munmap(file_memory, file_size) == -1) {
			perror("Error unmapping memory");
		}
		close(fd);
    }
    

	file_handle->write_file_hdr();
}