/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL v2.
You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */

#include "lock_manager.h"

/**
 * @description: 申请行级共享锁
 * @return {bool} 加锁是否成功
 * @param {Transaction*} txn 要申请锁的事务对象指针
 * @param {Rid&} rid 加锁的目标记录ID 记录所在的表的fd
 * @param {int} tab_fd
 */
bool LockManager::lock_shared_on_record(Transaction* txn, const Rid& rid, int tab_fd) {
    
    return true;
}

/**
 * @description: 申请行级排他锁
 * @return {bool} 加锁是否成功
 * @param {Transaction*} txn 要申请锁的事务对象指针
 * @param {Rid&} rid 加锁的目标记录ID
 * @param {int} tab_fd 记录所在的表的fd
 */
bool LockManager::lock_exclusive_on_record(Transaction* txn, const Rid& rid, int tab_fd) {

    return true;
}

/**
 * @description: 申请表级读锁
 * @return {bool} 返回加锁是否成功
 * @param {Transaction*} txn 要申请锁的事务对象指针
 * @param {int} tab_fd 目标表的fd
 */
bool LockManager::lock_shared_on_table(Transaction* txn, int tab_fd) {
    //先把锁表锁住
    std::unique_lock<std::mutex> lock{ latch_ };

    //然后查看事务状态
    if(txn->get_state()==TransactionState::SHRINKING){
        txn->set_state(TransactionState::ABORTED);
        throw TransactionAbortException(txn->get_transaction_id(), AbortReason::LOCK_ON_SHIRINKING);
    }
    //设置事务的状态是增长
    txn->set_state(TransactionState::GROWING);
    //构造锁数据id
    LockDataId lock_table_data_id=LockDataId(tab_fd,LockDataType::TABLE);
    LockRequest* request = nullptr;
    //根据id查看对应的锁请求
    auto lockrequestqueue=lock_table_.find(lock_table_data_id);
    //如果请求队列是空
    if(lockrequestqueue==lock_table_.end()){
        //构造一个新的请求队列
        std::unique_ptr<LockRequestQueue> lock_req_que=std::make_unique<LockRequestQueue>();
        lock_req_que->group_lock_mode_=GroupLockMode::S;
        //将新请求添加进去
        lock_req_que->request_queue_.emplace_back(txn->get_transaction_id(),LockMode::SHARED,false,txn->get_start_ts());
        //可能会有问题？      
        lock_table_[lock_table_data_id]=std::move(lock_req_que);
        
        request = &(lock_table_[lock_table_data_id]->request_queue_.back());
    }else{
        //查看已有请求
        auto& req_que=lockrequestqueue->second->request_queue_;
        for(auto& it:req_que){
            if(it.txn_id_==txn->get_transaction_id()){
                //如果已经分配了
                if(it.granted_){
                    //查看已经分配了的锁的类型
                    if(it.lock_mode_>=LockMode::SHARED){
                        return true;
                    }else{
                        //现在只有读锁和写锁，好像不会有这种情况？
                    }
                }else{
                    it.lock_mode_=LockMode::SHARED>it.lock_mode_?LockMode::SHARED:it.lock_mode_;
                    request=&(it);
                }
            }else{
                //锁冲突,其他事务占有锁并且是写锁，读锁可以一起上
                if(it.granted_&&it.lock_mode_==LockMode::EXLUCSIVE&&it.txn_start_ts<txn->get_start_ts()){
                    txn->set_state(TransactionState::ABORTED);
                    throw TransactionAbortException(txn->get_transaction_id(),AbortReason::DEADLOCK_PREVENTION);
                }else{

                }
            }
        }
        if(request==nullptr){
            //加入请求队列
            req_que.emplace_back(txn->get_transaction_id(),LockMode::SHARED,false,txn->get_start_ts());
            request=&req_que.back();
        }
    }
    //判断是否能上锁
    while(true){//死循环，没上锁就等着
        auto& request_queue=lock_table_[lock_table_data_id]->request_queue_;
        //遍历请求队列
        bool can_lock=true;
        for(auto it=request_queue.begin();it!=request_queue.end();it++){
            //如果已经为这个数据分配了锁
            if(it->granted_){
                //如果id相等，则不存在冲突
                if(it->txn_id_==txn->get_transaction_id()){
                    return true;
                }else if(it->txn_id_!=txn->get_transaction_id()&&it->lock_mode_==LockMode::SHARED)//如果id不相等，并且上的锁是读锁，也不冲突
                {
                    if(request->lock_mode_==LockMode::EXLUCSIVE){//如果请求是写锁
                        if(txn->get_start_ts()>it->txn_start_ts){//如果目前事务优先级低
                            txn->set_state(TransactionState::ABORTED);
                            throw TransactionAbortException(txn->get_transaction_id(),AbortReason::DEADLOCK_PREVENTION);
                        }
                        can_lock=false;
                    }
                    else{//读锁可以直接加
                        can_lock=true;
                    }
                    break;
                }else if(it->txn_id_!=txn->get_transaction_id()&&
                        it->lock_mode_==LockMode::EXLUCSIVE)
                {
                    //目前事务优先级低
                    if(it->txn_start_ts<txn->get_start_ts()){
                        txn->set_state(TransactionState::ABORTED);
                        throw TransactionAbortException(txn->get_transaction_id(),AbortReason::DEADLOCK_PREVENTION);
                    }else{//否则可以等一等
                        can_lock=false;
                        break;
                    }
                }
            }else{//没有给事务分配锁，继续查看下一个
                
            }
        }
        if(can_lock){
            // std::cout<<"lock for:"<<request->txn_id_<<" x"<<std::endl;
            request->granted_=true;
            txn->get_lock_set()->insert(lock_table_data_id);
            return true;
        }
        //等待变量等待别的线程释放
        lock_table_[lock_table_data_id]->cv_.wait(lock);
    }
    return true;
}

/**
 * @description: 申请表级写锁
 * @return {bool} 返回加锁是否成功
 * @param {Transaction*} txn 要申请锁的事务对象指针
 * @param {int} tab_fd 目标表的fd
 */
bool LockManager::lock_exclusive_on_table(Transaction* txn, int tab_fd) {
    //先把锁表锁住
    std::unique_lock<std::mutex> lock{ latch_ };

    //然后查看事务状态
    if(txn->get_state()==TransactionState::SHRINKING){
        txn->set_state(TransactionState::ABORTED);
        throw TransactionAbortException(txn->get_transaction_id(), AbortReason::LOCK_ON_SHIRINKING);
    }
    //设置事务的状态是增长
    txn->set_state(TransactionState::GROWING);
    //构造锁数据id
    LockDataId lock_table_data_id=LockDataId(tab_fd,LockDataType::TABLE);
    LockRequest* request = nullptr;
    //根据id查看对应的锁请求
    auto lockrequestqueue=lock_table_.find(lock_table_data_id);
    //如果请求队列是空
    if(lockrequestqueue==lock_table_.end()){
        //构造一个新的请求队列
        std::unique_ptr<LockRequestQueue> lock_req_que=std::make_unique<LockRequestQueue>();
        lock_req_que->group_lock_mode_=GroupLockMode::X;
        //将新请求添加进去
        lock_req_que->request_queue_.emplace_back(txn->get_transaction_id(),LockMode::EXLUCSIVE,false,txn->get_start_ts());
        //可能会有问题？      
        lock_table_[lock_table_data_id]=std::move(lock_req_que);
        
        request = new LockRequest(txn->get_transaction_id(),LockMode::EXLUCSIVE,false,txn->get_start_ts());
    }else{
        //查看已有请求
        auto& req_que=lockrequestqueue->second->request_queue_;
        lockrequestqueue->second->group_lock_mode_=GroupLockMode::X;        
        for(auto& it:req_que){
            if(it.txn_id_==txn->get_transaction_id()){
                //如果已经分配了
                if(it.granted_){
                    //查看已经分配了的锁的类型
                    if(it.lock_mode_==LockMode::EXLUCSIVE){
                        return true;
                    }else{
                        //如果当前的锁是s锁的话，升级请求
                        it.granted_=false;
                        it.lock_mode_=LockMode::EXLUCSIVE;
                        request=&(it);
                        
                    }
                }else{
                    it.lock_mode_=LockMode::EXLUCSIVE>it.lock_mode_?LockMode::EXLUCSIVE:it.lock_mode_;
                    request=&(it);
                }
            }else{
                //锁冲突,其他事务占有锁,并且该事务优先级更低，不需要请求了直接回滚
                if(it.granted_&&it.txn_start_ts<txn->get_start_ts()){
                    txn->set_state(TransactionState::ABORTED);
                    throw TransactionAbortException(txn->get_transaction_id(),AbortReason::DEADLOCK_PREVENTION);
                }else{

                }
            }
        }
        if(request==nullptr){
            //加入请求队列
            req_que.emplace_back(txn->get_transaction_id(),LockMode::EXLUCSIVE,false,txn->get_start_ts());
            request=&req_que.back();
        }
    }
    //判断是否能上锁
    while(true){//死循环，没上锁就等着
        auto request_queue=lock_table_[lock_table_data_id]->request_queue_;
        //遍历请求队列
        bool can_lock=true;
        for(auto it=request_queue.begin();it!=request_queue.end();it++){
            //如果已经为这个数据分配了锁
            if(it->granted_){
                //如果id相等，则不存在冲突
                if(it->txn_id_==txn->get_transaction_id()){
                    return true;
                }else if(it->txn_id_!=txn->get_transaction_id())
                {
                    //目前事务优先级低
                    if(it->txn_start_ts<txn->get_start_ts()){
                        txn->set_state(TransactionState::ABORTED);
                        throw TransactionAbortException(txn->get_transaction_id(),AbortReason::DEADLOCK_PREVENTION);
                    }else{//否则可以等一等
                        can_lock=false;
                        break;
                    }
                }
            }else{//没有给事务分配锁，继续查看下一个
                
            }
        }
        if(can_lock){
            request->granted_=true;
            txn->get_lock_set()->insert(lock_table_data_id);
            return true;
        }
        //等待变量等待别的线程释放
        lock_table_[lock_table_data_id]->cv_.wait(lock);
    }
    return true;
}

/**
 * @description: 申请表级意向读锁
 * @return {bool} 返回加锁是否成功
 * @param {Transaction*} txn 要申请锁的事务对象指针
 * @param {int} tab_fd 目标表的fd
 */
bool LockManager::lock_IS_on_table(Transaction* txn, int tab_fd) {
    
    return true;
}

/**
 * @description: 申请表级意向写锁
 * @return {bool} 返回加锁是否成功
 * @param {Transaction*} txn 要申请锁的事务对象指针
 * @param {int} tab_fd 目标表的fd
 */
bool LockManager::lock_IX_on_table(Transaction* txn, int tab_fd) {
    
    return true;
}

/**
 * @description: 释放锁
 * @return {bool} 返回解锁是否成功
 * @param {Transaction*} txn 要释放锁的事务对象指针
 * @param {LockDataId} lock_data_id 要释放的锁ID
 */
bool LockManager::unlock(Transaction* txn, LockDataId lock_data_id) {
    //保护锁表并发
    std::scoped_lock lock{latch_};
    //修改事务状态
    txn->set_state(TransactionState::SHRINKING);
    //如果锁表中不存在该对象，则直接返回
    if(!lock_table_.count(lock_data_id)){
        return false;
    }
    //删除锁请求队列中所出现的锁请求
    auto &request_queue=lock_table_[lock_data_id];
    bool can_erase=true;
    auto request=request_queue->request_queue_.begin();
    while(request!=request_queue->request_queue_.end()){
        if(request->txn_id_==txn->get_transaction_id()){
            request=request_queue->request_queue_.erase(request);
        }else{
            can_erase=false;
            request++;
        }
    }
    //释放当前锁id对应的资源
    if(can_erase)
    {
        lock_table_.erase(lock_data_id);
    }
    else{
        lock_table_[lock_data_id]->cv_.notify_all();
    }
    
    return true;
}